/****************************************************************************
 *
 * omp-simd-traffic.c - Biham-Middleton-Levine traffic model, OpenMP+SIMD version
 *
 * To cite this work, see the file CITATION
 *
 * Copyright (C) 2017, 2018, 2019 by Moreno Marzolla <moreno.marzolla(at)unibo.it>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/
#include "hpc.h"
#include <stdio.h>
#include <stdlib.h>     /* for rand(), srand()  */
#include <xmmintrin.h>
#include <assert.h>
#include <string.h>     /* for memcpy() */

typedef unsigned char cell_t;

typedef char v16i __attribute__((vector_size(16)));
#define VLEN (sizeof(v16i)/sizeof(char))

/* Possible values stored in a grid cell */
enum {
    EMPTY = 0,  /* empty cell                   */
    LR,         /* left-to-right moving vehicle */
    TB          /* top-to-bottom moving vehicle */
};

#define ISTART 1
#define IEND (n+1)
#define JSTART 1
#define JEND (n+1)

#define IDX(i,j) ((i)*(n+2) + (j))

void copy_top_bottom(cell_t *grid, int n)
{
    const int i = 1;
    memcpy(grid + IDX(n+i, JSTART), grid + IDX(i, JSTART), (JEND - JSTART)*sizeof(grid[0]));
    memcpy(grid + IDX(i-1, JSTART), grid + IDX(i-1+n, JSTART), (JEND - JSTART)*sizeof(grid[0]));
}

void copy_left_right(cell_t *grid, int n)
{
    int i;
    const int j = 1;
    for (i=ISTART; i<IEND; i++) {
        grid[IDX(i, n+j)] = grid[IDX(i,     j)]; /* left column to right halo */
        grid[IDX(i, j-1)] = grid[IDX(i, j-1+n)]; /* right column to left halo */
    }
}

v16i h_rule( v16i left, v16i center, v16i right )
{
    const v16i mask_lr = ((left == LR) & (center == EMPTY));
    const v16i mask_empty = ((center == LR) & (right == EMPTY));
    const v16i mask_center = ~(mask_lr | mask_empty);
    const v16i result = ((mask_lr & LR) |
                         (mask_empty & EMPTY) |
                         (mask_center & center));
    return result;
}

/* Move all left-to-right vehicles that are not blocked */
void horizontal_step( cell_t *cur, cell_t *next, int n )
{
    int i, j;
#pragma omp parallel for default(none) shared(cur,next,n) private(j)
    for (i=ISTART; i<IEND; i++) {
        const cell_t *pleft   = &cur[IDX(i,JSTART-1)];
        const cell_t *pcenter = &cur[IDX(i,  JSTART)];
        const cell_t *pright  = &cur[IDX(i,JSTART+1)];
        cell_t *pout = &next[IDX(i,JSTART)];
        for (j=JSTART; j<JEND-VLEN+1; j += VLEN) {
            const v16i left = __builtin_ia32_loaddqu((char*)pleft);
            const v16i center = __builtin_ia32_loaddqu((char*)pcenter);
            const v16i right = __builtin_ia32_loaddqu((char*)pright);
            const v16i result = h_rule(left, center, right);
            __builtin_ia32_storedqu((char*)pout, result);
            pleft += VLEN;
            pcenter += VLEN;
            pright += VLEN;
            pout += VLEN;
        }
    }
}

v16i v_rule( v16i top, v16i center, v16i bottom )
{
    const v16i mask_tb = ((top == TB) & (center == EMPTY));
    const v16i mask_empty = ((center == TB) & (bottom == EMPTY));
    const v16i mask_center = ~(mask_tb | mask_empty);
    const v16i result = ((mask_tb & TB) |
                         (mask_empty & EMPTY) |
                         (mask_center & center));
    return result;
}

/* Move all top-to-bottom vehicles that are not blocked */
void vertical_step( cell_t *cur, cell_t *next, int n )
{
    int i, j;
#pragma omp parallel for default(none) shared(cur,next,n) private(j)
    for (i=ISTART; i<IEND; i++) {
        const cell_t *ptop    = &cur[IDX(i-1,JSTART)];
        const cell_t *pcenter = &cur[IDX(i  ,JSTART)];
        const cell_t *pbottom = &cur[IDX(i+1,JSTART)];
        cell_t *pout = &next[IDX(i,JSTART)];
        for (j=JSTART; j<JEND-VLEN+1; j += VLEN) {
            const v16i top = __builtin_ia32_loaddqu((char*)ptop);
            const v16i center = __builtin_ia32_loaddqu((char*)pcenter);
            const v16i bottom = __builtin_ia32_loaddqu((char*)pbottom);
            const v16i result = v_rule(top, center, bottom);
            __builtin_ia32_storedqu((char*)pout, result);
            ptop += VLEN;
            pcenter += VLEN;
            pbottom += VLEN;
            pout += VLEN;
        }
    }
}

/* Returns 1 with probability p */
int rand01( float p )
{
    return ((rand() / (float)RAND_MAX) < p);
}

/* Initialize |grid| with vehicles with density |rho|. The direction of
   each vehicle is chosen with equal probability. */
void setup( cell_t* grid, int n, float rho )
{
    int i, j;
    for ( i=ISTART; i<IEND; i++ ) {
        for ( j=JSTART; j<JEND; j++) {
            if ( rand01( rho ) ) {
                grid[IDX(i,j)] = ( rand01(0.5) ? LR : TB );
            } else {
                grid[IDX(i,j)] = EMPTY;
            }
        }
    }
}

/* Dump |grid| as a PPM (Portable PixMap) image written to file
   |filename|. LR vehicles are shown as blue pixels, while TB vehicles
   are shown in red. Empty cells are white. */
void dump( const cell_t *grid, int n, const char* filename )
{
    int i, j;
    FILE *out = fopen( filename, "w" );
    if ( NULL == out ) {
        printf("Cannot create file %s\n", filename);
        abort();
    }
    fprintf(out, "P6\n");
    fprintf(out, "%d %d\n", n, n);
    fprintf(out, "255\n");
    for (i=ISTART; i<IEND; i++) {
        for (j=JSTART; j<JEND; j++) {
            switch( grid[IDX(i,j)] ) {
            case EMPTY:
                fprintf(out, "%c%c%c", 255, 255, 255);
                break;
            case TB:
                fprintf(out, "%c%c%c", 0, 0, 255);
                break;
            case LR:
                fprintf(out, "%c%c%c", 255, 0, 0);
                break;
            default:
                printf("Error: unknown cell state %u\n", grid[IDX(i,j)]);
                abort();
            }
        }
    }
    fclose(out);
}

#define BUFLEN 256

int main( int argc, char* argv[] )
{
    cell_t *cur, *next;
    char buf[BUFLEN];
    const char* pattern = "omp-simd-traffic-%05d.ppm";
    int s, N = 256, nsteps = 1024;
    float rho = 0.2;
    double tstart, tend;

    if ( argc > 4 ) {
        printf("Usage: %s [nsteps [rho [N]]]\n", argv[0]);
        return EXIT_FAILURE;
    }

    if ( argc > 1 ) {
        nsteps = atoi(argv[1]);
    }

    if ( argc > 2 ) {
        rho = atof(argv[2]);
    }

    if ( argc > 3 ) {
        N = atoi(argv[3]);
    }

    assert( 0 == N % VLEN );

    const size_t size = (N+2)*(N+2)*sizeof(cell_t);

    /* Allocate grids */
    cur = (cell_t*)malloc(size);
    next = (cell_t*)malloc(size);

    setup(cur, N, rho);
    tstart = hpc_gettime();

    for (s=0; s<nsteps; s++) {
#ifdef DUMPALL
        snprintf(buf, BUFLEN, pattern, s);
        dump(cur, N, buf);
#endif
        copy_left_right(cur, N);
        horizontal_step(cur, next, N);
        copy_top_bottom(next, N);
        vertical_step(next, cur, N);
    }
    tend = hpc_gettime();
    const double elapsed = tend - tstart;
    const double nupdates = 1.0e-6 * N * N * nsteps;
    fprintf(stderr, "%s : %f Mupdates in %.4f seconds (%f Mupd/sec)\n", argv[0], nupdates, elapsed, nupdates/elapsed);

    /* dump last state */
    snprintf(buf, BUFLEN, pattern, s);
    dump(cur, N, buf);

    /* Free memory */
    free(cur);
    free(next);
    return EXIT_SUCCESS;
}
